package fractal;

class Grammar
{
    final static int MAX_STRING = 1000000;

    byte[] strBytes_ = new byte[MAX_STRING];
    ListCell start_;
    ListCell[] rules_ = new ListCell[256];

    class ListCell 
      {
        public int char_;
        public ListCell next_;
        
        public ListCell(int aChar, ListCell next)
          {
            char_ = aChar;
            next_ = next;
          }
      }
    
    public Grammar(String[] rules) throws InvalidRule
      {
        start_ = new ListCell((int)rules[0].charAt(0), null);
        
        for(int i = 1; i < rules.length; i++)
            readRule(rules[i]);
      }
    
    void readRule(String rule) throws InvalidRule
      {
        int leftSide = (int)rule.charAt(0);
        if(rule.length() < 3)
            throw new InvalidRule(rule);
        if((rule.charAt(1) != '-') || 
            (rule.charAt(2) != '>'))
            throw new InvalidRule(rule);
            
        ListCell curr = null;
        for(int i = rule.length() - 1; i >= 3; i--)
          {
            curr = new ListCell((int)rule.charAt(i), curr);
          }
          
        rules_[leftSide] = curr;
      }
      
    String nextStep()
      {
        ListCell curr = start_, prev = null;
        while(curr != null)
          {
            if(rules_[curr.char_] != null) // make substitution with right side
              {
                ListCell first = dupList(rules_[curr.char_]);
                ListCell last = lastCell(first);
                last.next_ = curr.next_;
                if(prev != null)
                    prev.next_ = first;
                else
                    start_ = first;
                curr = last;
              }
            prev = curr;
            curr = curr.next_;
          }
        return listToString(start_);
      }
      
    ListCell dupList(ListCell list)
      {
        ListCell retval = null, curr = list, currcpy = null;
        while(curr != null)
          {
            if(currcpy == null)
                retval = currcpy = new ListCell(curr.char_, null);
            else
              {
                currcpy.next_ = new ListCell(curr.char_, null);
                currcpy = currcpy.next_;
              }
            curr = curr.next_;
          }
        return retval;
      }
      
    ListCell lastCell(ListCell list)
      {
        ListCell last = list;
        if(last == null)
            return null;
        while(last.next_ != null)
            last = last.next_;
        return last;
      }

    String listToString(ListCell list)
      {
        int length = 0;
        ListCell curr = list;
        while(curr != null)
          {
            strBytes_[length++] = (byte)curr.char_;
            curr = curr.next_;
          }
        return new String(strBytes_, 0, length);
      }

    public static void main(String[] args) throws InvalidRule
      {
        if(args.length < 2)
          {
            System.out.println("usage: java Grammar <start> <rule1> [<rule2> <rule3> ...]");
            System.exit(1);
          }

        String step = null;
        Grammar g = new Grammar(args);
        for(int i = 0; i < 3; i++)
            System.out.println(step = g.nextStep());

        Model3D m = new Model3D(1, 2 * Math.PI * 20 / 360, step);
        System.out.println("The model is:\n" + m);
      }
}
