package fractal;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

// implements Lindenmayer plant fractals
public class Lm extends Applet
{
    public static final int WIDTH = 900, HEIGHT = 750;
    public static final int INITIAL_N = 3;
    public static final double INITIAL_D = 10;
    public static final double INITIAL_ANGLE = 18.0;
    
    public final static String GRAMMAR_1_TITLE = "tree 2D 1";
    public final static String GRAMMAR_1 = "F\nF->FF-[-F+F+F]+[+F-F-F]\n";
    public final static String[] GRAMMAR_1_RULES = 
        {"F", "F->FF-[-F+F+F]+[+F-F-F]"};

    public final static String GRAMMAR_2_TITLE = "tree 2D 2";
    public final static String GRAMMAR_2 = 
        "X\nX->F-[[X]+X]+F[+FX]-X\nF->FF\n";
    public final static String[] GRAMMAR_2_RULES = 
        {"X", "X->F-[[X]+X]+F[+FX]-X", "F->FF"};

    public final static String GRAMMAR_3_TITLE = "bush 3D";
    public final static String GRAMMAR_3 = 
        "O\nO->//'''''''''''''ta\nt->?t\na->[[&sl!a]/////'[&sl!`a]///////'[&sl!``a]]\ns->Sl\nS->S/////s\nl->['''&&{-f+f+f-|-f+f+f}]\n";
    public final static String[] GRAMMAR_3_RULES = 
        {"O", 
         "O->??????????//'''''''''''''ta",
         "t->?t",
         "a->[[&sl!a]/////'[&sl!`a]///////'[&sl!``a]]", 
         "s->Sl", 
         "S->S/////s", 
         "l->['''&&{-f+f+f-|-f+f+f}]"};

    public final static String GRAMMAR_4_TITLE = "flower 3D 1";
    public final static String GRAMMAR_4 = 
        "p\np->i+[p+o]--//[--l]i[++l]-[po]++po\ni->Fs[//&&l][//^^l]Fs\ns->sFs\nl->[''''''{+f-ff-F+|+f-ff-f}]\no->[&&&d~/w////w////w////w////w]\nd->FF\n";
    public final static String[] GRAMMAR_4_RULES = 
        {"p",
        "p->i+[p+o]--//[--l]i[++l]-[po]++po",
        "i->Fs[//&&l][//^^l]Fs",
        "s->sFs",
        "l->[''''''{+f-ff-f+|+f-ff-f}]",
        "o->[&&&d~/w////w////w////w////w]",
        "d->FF"};

    public final static String GRAMMAR_5_TITLE = "flower 3D 2";
    public final static String GRAMMAR_5 = 
        "O\nO->''''''''''''''p\np->i+[p+o]--//[--l]i[++l]-[po]++po\ni->Fs[//&&l][//^^l]Fs\ns->sFs\nl->[''''{+f-ff-f+|+f-ff-f}]\no->[&&&d~/w////w////w////w////w]\nd->FF\nw->[^``````F][````{&&&&-f+f|-f+f}]";
    public final static String[] GRAMMAR_5_RULES = 
        {"O",
        "O->''''''''''''''p",
        "p->i+[p+o]--//[--l]i[++l]-[po]++po",
        "i->Fs[//&&l][//^^l]Fs",
        "s->sFs",
        "l->[''''{+f-ff-f+|+f-ff-f}]",
        "o->[&&&d~/w////w////w////w////w]",
        "d->FF",
        "w->[^``````F][````{&&&&-f+f|-f+f}]"};

    public final static String GRAMMAR_6_TITLE = "fir tree 2D";
    public final static String GRAMMAR_6 = 
        "O\nO->''''''''''''''''''ULFFF\nU->[+++G][---G]TU\nG->+H[-G]L\nH->-G[+H]L\nT->TL\nL->[-FFF][+FFF]F";
    public final static String[] GRAMMAR_6_RULES = 
        {"O",
        "O->''''''''''''''''''ULFFF",
        "U->[+++G][---G]TU",
        "G->+H[-G]L",
        "H->-G[+H]L",
        "T->TL",
        "L->[-FFF][+FFF]F"};


    
    public static final String[]
        defaultGrammarTitles_ = {GRAMMAR_1_TITLE, 
                                GRAMMAR_2_TITLE, 
                                GRAMMAR_3_TITLE,
                                GRAMMAR_4_TITLE,
                                GRAMMAR_5_TITLE,
                                GRAMMAR_6_TITLE
                                },
        defaultGrammars_ = {GRAMMAR_1,
                            GRAMMAR_2,
                            GRAMMAR_3,
                            GRAMMAR_4,
                            GRAMMAR_5,
                            GRAMMAR_6
                            };

    int n_ = INITIAL_N;
    double d_ = INITIAL_D, angle_ = INITIAL_ANGLE;
    ImageCanvas fractalCanvas_;
    boolean modified_ = true;

    Button refreshB_, prevB_, nextB_;
    TextField nTF_, angleTF_, dTF_;
    TextArea grammarTA_;
    Choice grammarC_;
    String[] rules_ = GRAMMAR_1_RULES;
    
    public Lm()
      {
        this.setSize(WIDTH, HEIGHT);

        refreshB_ = new Button("Draw");
        refreshB_.addActionListener(new refreshBListener());
        nextB_ = new Button("Next");
        nextB_.addActionListener(new nextBListener());
        prevB_ = new Button("Prev");
        prevB_.addActionListener(new prevBListener());
        Label nL = new Label("# iterations");
        Label dL = new Label("step");
        Label angleL = new Label("angle");
        nTF_ = new TextField("" + INITIAL_N, 3);
        angleTF_ = new TextField("" + INITIAL_ANGLE, 6);
        dTF_ = new TextField("" + INITIAL_D, 6);
        grammarTA_ = new TextArea(defaultGrammars_[0],
            8, 40, TextArea.SCROLLBARS_BOTH);        

        grammarC_ = new Choice();
        for(int i = 0; i < defaultGrammarTitles_.length; i++)
            grammarC_.add(defaultGrammarTitles_[i]);
        grammarC_.select(0);
        grammarC_.addItemListener(new grammarCListener());
        fractalCanvas_ = new ImageCanvas(this);

        this.add(refreshB_);
        this.add(nextB_);
        this.add(prevB_);
        this.add(nL);
        this.add(nTF_);
        this.add(dL);
        this.add(dTF_);
        this.add(angleL);
        this.add(angleTF_);
        this.add(grammarC_);
        this.add(grammarTA_);
        this.add(fractalCanvas_);
      }

    public void init()
      {
//        this.getParent().validate();
      }

  public void update(Graphics g) {
    paint(g);
  }
    
    static class KillAdapter extends WindowAdapter
      {
        public void windowClosing(WindowEvent event)
          {
            System.exit(0);
          }
      }

    public static void main(String[] args)
      {
        Lm lm = new Lm();

        // create a frame for it
        Frame lmFrame = new Frame("Lindenmayer Program");

        //set the size of the frame
        lmFrame.setSize(WIDTH + 20, HEIGHT + 10);

        //add the image to the frame
        lmFrame.add(lm);

        // make the window closeable
        lmFrame.addWindowListener(new Lm.KillAdapter());

        // show it up on the screen
        lmFrame.setVisible(true);
      }

    class refreshBListener implements ActionListener
      {
        public void actionPerformed(ActionEvent e)
          {
            n_ = Integer.parseInt(nTF_.getText());
            d_ = new Double(dTF_.getText()).doubleValue(); 
            angle_ = new Double(angleTF_.getText()).doubleValue();
            modified_ = true;
            String rulesS = grammarTA_.getText();
            StringTokenizer st = new StringTokenizer(rulesS, "\n");
            rules_ = new String[st.countTokens()];
            for(int i = 0; i < rules_.length; i++)
                rules_[i] = st.nextToken();
            fractalCanvas_.repaint();
          }
      }
      
    class nextBListener implements ActionListener
      {
        public void actionPerformed(ActionEvent e)
          {
            n_ = Integer.parseInt(nTF_.getText());
            nTF_.setText("" + ++n_);
            d_ = new Double(dTF_.getText()).doubleValue(); 
            angle_ = new Double(angleTF_.getText()).doubleValue();
            modified_ = true;
            String rulesS = grammarTA_.getText();
            StringTokenizer st = new StringTokenizer(rulesS, "\n");
            rules_ = new String[st.countTokens()];
            for(int i = 0; i < rules_.length; i++)
                rules_[i] = st.nextToken();
            fractalCanvas_.repaint();
          }
      }
      
    class prevBListener implements ActionListener
      {
        public void actionPerformed(ActionEvent e)
          {
            n_ = Integer.parseInt(nTF_.getText());
            if(n_ > 0)
              {
                nTF_.setText("" + --n_);
                d_ = new Double(dTF_.getText()).doubleValue(); 
                angle_ = new Double(angleTF_.getText()).doubleValue();
                modified_ = true;
                String rulesS = grammarTA_.getText();
                StringTokenizer st = new StringTokenizer(rulesS, "\n");
                rules_ = new String[st.countTokens()];
                for(int i = 0; i < rules_.length; i++)
                    rules_[i] = st.nextToken();
                fractalCanvas_.repaint();
              }
          }
      }
      
    class grammarCListener implements ItemListener
      {
        public void itemStateChanged(ItemEvent e)
          {
            int index = grammarC_.getSelectedIndex();
            grammarTA_.setText(defaultGrammars_[index]);
          }
      }
}
