package fractal;

class Model3D
{
    final static int MAX_STACK = 200;
    final static int MAX_POLY  = 20;
    final static double TH_STEP  = 0.2;

    Object3DCell model_ = new Object3DCell(null, null);
    double[][] hplus = new double[3][3],
        hminus = new double[3][3], 
        lplus = new double[3][3], 
        lminus = new double[3][3], 
        uplus = new double[3][3], 
        uminus = new double[3][3], 
        uback = new double[3][3];

    public Model3D(double d, double angle, String turtleString)
      {
        double[][] orientation = {{1, 0, 0}, {0, 1, 0}, {0, 0, 1}};
        double[] position_ = {0, 0, 0};
        int color_ = 0;
        double thickness_ = 0;
        Object3DCell current = model_;
        
        initRotationMatrices(angle);
        
        double[] last = new double[3];
        int i, lin, col, sp = 0;
        double[] newvector = null;
        double[][] newmatrix = null;
        Object[] state = new Object[MAX_STACK];
        
        double[][] poly = new double[MAX_POLY][];
        int polyCount = 3;
        
        for(i = 0; i < turtleString.length(); i++)
          {
            switch(turtleString.charAt(i))
              {
                case 'S': // line segments
                case 'F':
                    position_[0] += d * orientation[0][0];
                    position_[1] += d * orientation[1][0];
                    position_[2] += d * orientation[2][0];
//                    current = current.next_ = new Object3DCell(
//                                new Line3D(last, position_, color_, thickness_), 
//                                null);
                    current.next_ = new Object3DCell(
                                new Line3D(last, position_, color_, thickness_), 
                                current.next_);
                    last[0] = position_[0];
                    last[1] = position_[1];
                    last[2] = position_[2];
                    break;
                case 'f': // line segments that are part of a polygon
                    poly[polyCount] = new double[3];
                    poly[polyCount][0] = (position_[0] += d * orientation[0][0]);
                    poly[polyCount][1] = (position_[1] += d * orientation[1][0]);
                    poly[polyCount][2] = (position_[2] += d * orientation[2][0]);
                    polyCount++;
                    last[0] = position_[0];
                    last[1] = position_[1];
                    last[2] = position_[2];
                    break;
                case '-':
                    rotate(uminus, orientation);
                    break;
                case '+':
                    rotate(uplus, orientation);
                    break;
                case '^':
                    rotate(lplus, orientation);
                    break;
                case '&':
                    rotate(lminus, orientation);
                    break;
                case '\\': // roll left
                    rotate(hplus, orientation);
                    break;
                case '/': // roll right
                    rotate(hminus, orientation);
                    break;
                case '|': // turn around 180 degrees
                    rotate(uback, orientation);
                    break;
                case '\'': // increase color
                    color_++;
                    break;
                case '`': // decrease color
                    color_--;
                    break;
                case '?': // increase thickness
                    thickness_ += TH_STEP;
                    break;
                case '!': // decrease thickness
                    thickness_ -= TH_STEP;
                    break;
                case '[': // push state on stack
                    newmatrix = new double[3][3];
                    for(lin = 0; lin < 3; lin++)
                        for(col = 0; col < 3; col++)
                            newmatrix[lin][col] = orientation[lin][col];

                    newvector = new double[3];
                    for(lin = 0; lin < 3; lin++)
                        newvector[lin] = position_[lin];
                    state[sp++] = newvector;
                    state[sp++] = newmatrix;
                    state[sp++] = new Integer(color_);
                    state[sp++] = new Double(thickness_);
                    break;
                case ']': // pop state from stack
                    thickness_ = ((Double)state[--sp]).doubleValue();
                    color_ = ((Integer)state[--sp]).intValue();
                    newmatrix = (double[][])state[--sp];
                    state[sp] = null; // careful, let it garbage collect
                    newvector = (double[])state[--sp];
                    state[sp] = null; // careful, let it garbage collect
                    
                    for(lin = 0; lin < 3; lin++)
                        for(col = 0; col < 3; col++)
                            orientation[lin][col] = newmatrix[lin][col];

                    for(lin = 0; lin < 3; lin++)
                        last[lin] = position_[lin] = newvector[lin];
                    break;
                case '{': // begin polygon
                    polyCount = 0;
                    break;
                case '}': // end polygon
//                    current = current.next_ = new Object3DCell(
//                        new Polygon3D(poly, polyCount, color_), null);
                    current.next_ = new Object3DCell(
                        new Polygon3D(poly, polyCount, color_), current.next_);
                    break;
              }
          }
      }

    void rotate(double[][] rot, double[][] orientation)
      {
        double[][] aux = new double[3][3];
        int lin, col, k;
        for(lin = 0; lin < 3; lin++)
            for(col = 0; col < 3; col++)
              {
                aux[lin][col] = 0;
                for(k = 0; k < 3; k++)
                    aux[lin][col] += orientation[lin][k] * rot[k][col];
              }
        for(lin = 0; lin < 3; lin++)
            for(col = 0; col < 3; col++)
                orientation[lin][col] = aux[lin][col];
      }

    void initRotationMatrices(double angle)
      {
        lminus[2][2] = lminus[0][0] = lplus[2][2] = lplus[0][0] = 
        hminus[1][1] = hminus[2][2] = hplus[1][1] = hplus[2][2] = 
        uminus[0][0] = uminus[1][1] = uplus[0][0] = uplus[1][1] = Math.cos(angle);

        lplus[2][0] = lminus[0][2] =
        hminus[2][1] = hplus[1][2] =
        uminus[1][0] = uplus[0][1] = 
        -(lplus[0][2] = lminus[2][0] =
        hplus[2][1] = hminus[1][2] =
        uplus[1][0] = uminus[0][1] = Math.sin(angle));

        lminus[1][1] = lplus[1][1] = 
        hminus[0][0] = hplus[0][0] = 
        uminus[2][2] = uplus[2][2] = 1;
        
        uback[0][0] = uback[1][1] = -(uback[2][2] = 1);
      }
      
    Object3DCell getModel()
      {
        return model_.next_;
      }
      
    public String toString()
      {
        Object3DCell curr = model_;
        int len = 0;
        String retval = "";
        
        for(;curr != null; curr = curr.next_)
          {
            retval = " " + curr + retval;
            len++;
          }
        return "# objs: " + len + " " + retval;
      }
}
